/* ump3c.c
**
** ump3c: Copyright 2002 Christian Platt
** licensed under the GNU General Public License (GPL)
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
** See the included COPYING file for the license.
** 
** This program uses a lot of code from the dist10 mp3 decoder by the ISO
** MPEG Audio Subgroup Software Simulation Group.  See below for more
** details.
**
**
** 7/20/02
**
** This file contains the main function for the UnderMP3Cover program (ump3c).
** It also contains the hiding and reading functions that are called in the
** decode.c file when decoding the mp3.  The main, usage, and getArguments 
** routines are highly modified versions of routines that were originally 
** found in the musicout.c file for the dist10 mp3 decoder by MP3 Decoder
** (dist10) of the ISO MPEG Audio Subgroup Software Simulation Group. All 
** other files and code are from this dist10 decoder but some have been modifed
** for ump3c.  The modification are commented as such. The files that have
** been modified for ump3c are:
**       common.c
**       common.h
**       decode.c
**
**
** All other files (whether they are actually used or not) have been left in
** the building process of this program since there are future plans of
** more developemnt of UnderMP3Cover that may need to use some code in these
** files.
**
** For instructions on how to use ump3c, type in "ump3c" or "ump3c -help"
** at the command prompt.
**
** 4/10/04
**
** Added verbose mode, via the -v command line flag. Now no output statements
** are given (excpet error messages) unless this flag is specified.
**
*/


#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include "common.h"
#include "decoder.h"

#define DEFAULT_SPACING   2    /* default spacing of hidden bits in the mp3 */

/* functions definitions */
void usage();
void getArguments();

/* these functions are called from decode.c */
int hideFileSize(unsigned * info);
int hideByte(unsigned * info);
int readHiddenByte(unsigned info);
int readFileSize(unsigned info);  

/* Global variable definitions for "musicout.c" */
int main_data_slots();


/* global variables for ump3c */
FILE * ptrHiddenFile;         /* ptr to the hidden file being hidde/created */
Bit_stream_struc  ptrNewMP3;  /* ptr to new mp3 file */
int hiding;                   /* flag used to signal hiding or reading */
int bitSpacing;               /* spacing of hiding an reading bits from mp3 */
unsigned int inFileSize;      /* size of file that is going to be hidden */
int tempFilePtr;              /* temp file ptr for finding file sizes */
int nBytesHidden = 0;         /* number of bytes that were hidden in the mp3 */
int nBytesRead = 0;           /* number of hidden bytes read from the mp3 */
int hiddenFileSize = 2;       /* size of file that is already hidden in mp3 */
                              /* used when extracting out hidden file */
                              /* it is initialized to 2 to mark that the */
                              /* most sig bit is a 1 */
int VERBOSE_MODE = 0;         /* flag for displaying output statements */


int main(int argc, char **argv) {
  Bit_stream_struc ptrOrigMP3;  /* ptr to orig mp3 */
  Arguments_t Arguments;        /* struct used for storing command line args */
  
  /* variables used by dist10 mp3 decoding */
  int i, sync;
  frame_params fr_ps;
  layer info;
  unsigned int old_crc;
  unsigned long bitsPerSlot, samplesPerFrame;
  unsigned long frameBits, gotBits = 0;
  III_side_info_t III_side_info;

  /* temp variables for transfering putbits */
  unsigned long tempLong;
  int tempInt;

  /* assign frame info */
  fr_ps.header = &info;
  fr_ps.tab_num = -1;   
  fr_ps.alloc = NULL;

  /* parse the command line arguments */
  getArguments(argc, argv, &Arguments);

  /* open or create the necessary files */
  /* see if we are in hiding mode */
  if (Arguments.hiding == 1) {
    /* set the hiding/reading flag */
    hiding = 1;

    /* open the file to be hidden  */
    /* first open up the file to get its size */
    tempFilePtr = open (Arguments.hiddenFileName, O_RDONLY);
    inFileSize = lseek (tempFilePtr, 0, SEEK_END);
    close (tempFilePtr);

    /* now open it to read bytes from it */
    ptrHiddenFile = fopen(Arguments.hiddenFileName, "rb");
    /* check for error when opening */
    if (!(ptrHiddenFile)) {
      printf ("Aborted!\n");
      printf ("Unable to open file to be hidden: %s\n",
               Arguments.hiddenFileName);
      exit(1);
    }

  /* we are in reading mode */
  } else if (Arguments.hiding == 0){
    /* set the hiding/reading flag */
    hiding = 0;

    /* open file to write the hidden file to */
    ptrHiddenFile = fopen(Arguments.hiddenFileName, "wb");
    /* check for error when opening */
    if (!(ptrHiddenFile)) {
      printf ("Aborted!\n");
      printf ("Unable to open file to write hidden file to: %s\n",
               Arguments.hiddenFileName);
      exit(1);
    }
  }

  /* set the bit spacing */
  bitSpacing = Arguments.bitSpacing;

  /* open original mp3 file to read from */
  /* if in reading mode then this is the mp3 that holds the hidden file */
  /* if in hiding mode then this is the orig mp3 to create new mp3 from */
  open_bit_stream_r(&ptrOrigMP3, Arguments.origMP3FileName, BUFFER_SIZE);

  /* if we are in hiding mode create new mp3 file */
  if (hiding) {
    open_bit_stream_w(&ptrNewMP3, Arguments.newMP3FileName, BUFFER_SIZE);
    if (VERBOSE_MODE == 1) {
      printf ("Hiding file...\n");
    }
  } else {
    if (VERBOSE_MODE == 1) {
      printf ("Reading hidden file...\n");
    }
  }

  /* begin looping through the mp3 files */
  while (!end_bs(&ptrOrigMP3)) {
    /* dist10 mp3 decoding variables */
    int nSlots;
    int main_data_end, flush_main;
    int bytes_to_discard;
    static int frame_start = 0;

    /* get synced up with the frames */
    sync = seek_sync(&ptrOrigMP3, SYNC_WORD, SYNC_WORD_LNGTH);
    frameBits = sstell(&ptrOrigMP3) - gotBits;
    gotBits += frameBits;

    /* make sure we are in sync */
    if (!sync) {
      break;
    }

    /* get information from the frame header */
    /* this routine will write some bytes to the new mp3 */
    /* if in hiding mode */
    decode_info(&ptrOrigMP3, &fr_ps);
    hdr_to_frps(&fr_ps);

    /* check for error protection */
    if (info.error_protection) {
      buffer_CRC(&ptrOrigMP3, &old_crc);
    }

    /* determine the number of samples per frame */
    bitsPerSlot = 8;        
    if(fr_ps.header->version == MPEG_PHASE2_LSF)
      samplesPerFrame = 576;
    else
      samplesPerFrame = 1152;

    /* read in side information */
    /* this is where the hidden file is inserted or extracted */ 
    III_get_side_info(&ptrOrigMP3, &III_side_info, &fr_ps);
    if (hiding) {
      III_put_side_info(&ptrNewMP3, &III_side_info, &fr_ps);
    }

    /* now read the main date for the frame */
    nSlots = main_data_slots(fr_ps);
    for (; nSlots > 0; nSlots--){  /* read main data. */
      tempLong = getbits(&ptrOrigMP3,8);
      hputbuf((unsigned int) tempLong, 8);
      /* need to write these bits to the new mp3 */
      if (hiding) {
        putbits(&ptrNewMP3, tempLong, 8);
      }
    }

    /* look for end of the frame */
    main_data_end = hsstell() / 8; /*of privious frame*/
    if ( flush_main=(hsstell() % bitsPerSlot) ) { 
      hgetbits((int)(bitsPerSlot - flush_main));
      main_data_end++;
    }
    bytes_to_discard = frame_start - main_data_end
                         - III_side_info.main_data_begin ;
    if(main_data_end > 4096) {
      frame_start -= 4096;
      rewindNbytes(4096);
    }

    frame_start += main_data_slots(fr_ps);
    for (; bytes_to_discard > 0; bytes_to_discard--) {
      hgetbits(8);
    }

    /* read in potential ancillary data */
    if (info.bitrate_index > 0) { /* if not free-format */
      long anc_len;
      anc_len = (int)((double)samplesPerFrame /
                s_freq[info.version][info.sampling_frequency] *
	        (double)bitrate[info.version][info.lay-1][info.bitrate_index] /
			       (double)bitsPerSlot);
      if (info.padding) {
        anc_len++;
      }
      anc_len *= bitsPerSlot;
      anc_len -= sstell(&ptrOrigMP3)-gotBits+SYNC_WORD_LNGTH;
      for (i=0; i<anc_len; i++) {
        tempInt = get1bit(&ptrOrigMP3);
        if (hiding) {
          put1bit(&ptrNewMP3, tempInt);
        }
      }
    }
  }

  /* output results */

  /* check for no transfer of data. This is most likely because of */
  /* problems with specified mp3s */
  if (hiding && nBytesHidden == 0)  {
    printf("Error!\n");
    printf("Either there were problems hiding data in specified mp3: %s\n", 
            Arguments.origMP3FileName);
    printf("Or the specified file to be hidden is empty: %s\n",
            Arguments.hiddenFileName);
  } else if (!(hiding) && nBytesRead == 0) {
    printf("Error! Problems reading data from specified mp3: %s\n", 
            Arguments.origMP3FileName);

  /* output success results */
  } else {

    /* if we made it this far then data was either hidden or read */
    if (VERBOSE_MODE == 1) {
      printf("Success!\n");
      if (hiding) {
        printf ("The file %s has been hidden into %s at a spacing of %d.\n",
                 Arguments.hiddenFileName,
                 Arguments.origMP3FileName,
                 bitSpacing);
        printf("(To extract the file you must use the same spacing value)\n");
      } else {
        printf ("The file %s has been extracted from %s at a spacing of %d.\n",
                 Arguments.hiddenFileName,
                 Arguments.origMP3FileName,
                 bitSpacing);
      }
    }

    /* check to see if all of the file was able to be hidden in the mp3 */
    if (hiding && (nBytesHidden < inFileSize)) {
      printf("Warning! This mp3 was not big enough to hide the entire file.\n");
      printf("Not all of the file was able to be hidden into the mp3.\n");
      printf("Try a smaller spacing or larger mp3 file.\n");
    } else if (!(hiding) && (nBytesRead < hiddenFileSize)) {
      printf("Warning! The extracted hidden file may be incomplete.\n");
      printf("The hidden file size was larger than what was extracted.\n");
      printf("If possible try hiding the file again.\n");
    } 
  }     

  /* close the files we opened */
  close_bit_stream_r(&ptrOrigMP3);
  if (hiding) {
    close_bit_stream_w(&ptrNewMP3);
  } 
  fclose(ptrHiddenFile);

  /* We need to get rid of the last 2 bytes in the new mp3 */
  /* This is becuase of the use of the buffer in reading the main data */
  /* some of this data gets discarded when reading to be read later, unless */
  /* we reach the end of the file. But when writing it at the end, all of */
  /* of the buffer gets written. Therefor we have 2 extra bytes of junk */
  /* at the end of the new mp3 */
  if (hiding && nBytesHidden != 0) {
    tempFilePtr = open (Arguments.newMP3FileName, O_RDONLY);
    inFileSize = lseek (tempFilePtr, 0, SEEK_END);
    close (tempFilePtr);
    if (truncate (Arguments.newMP3FileName, (inFileSize - 2)) != 0){
      printf("Error truncating new mp3 file %s.\n", 
                                        Arguments.newMP3FileName);
    }

  }

  exit(0);
}



/* usage()
**
** This routine outputs the usage screen for ump3c.  It describes
** how to use the program and what each variable represents.
** It is called when the user types in "ump3c -help" or just "ump3c".
*/
void usage() {
  printf ("usage: ump3c\n");
  printf ("\n");
  printf ("For hiding a file in an mp3:\n");
  printf ("\n");
  printf ("      ump3c -h [-v] [-c N] origMP3 modifiedMP3 hiddenFile\n");
  printf ("\n");
  printf ("For reading a hidden file from an mp3:\n");
  printf ("\n");
  printf ("      ump3c -r [-v] [-c N] modifiedMP3 newFile\n");
  printf ("\n");
  printf ("where\n");
  printf ("  -h           hiding file in mp3\n");
  printf ("  -r           reading hidden file from mp3\n");
  printf ("  -v           display output statements\n");
  printf ("  -c N         bit spacing of hidden file in mp3(N defaults to %d)\n", DEFAULT_SPACING);
  printf ("  origMP3      original mp3 file to create a new mp3 from (with\n");
  printf ("               the hidden file inside. origMP3 will not be modified)\n");
  printf ("  modifiedMP3  name of modified mp3 that contains the hidden file\n");
  printf ("  hiddenFile   name of file to be hidden in mp3\n");
  printf ("  newFile      name of hidden file that is extracted from mp3\n");
  printf ("  -help        to display this usage menu\n");
  printf ("\n");
  printf ("  NOTES: For -c N, N must be greater than 0.\n");
  printf ("\n");
  printf ("         For -c N, N must be the same value for hiding and\n");
  printf ("         reading to be able to retrieve the hidden file.\n");
  printf ("\n");
  printf ("         If the modifiedMP3 has some audible differences\n");
  printf ("         (ex. slight tremelo effect) try increasing N to see\n");
  printf ("         if this helps.  But remember, by increasing N you will\n");
  printf ("         be decreasing the maximum size of a file that can be\n");
  printf ("         hidden.\n");
  printf ("\n");
  printf ("         To increase the maximum file size that can be hidden in\n");
  printf ("         a particular mp3 decrease N (-c N). But remember, low values\n");
  printf ("         of N may create slight audible differences. This will\n");
  printf ("         depend on the particular mp3.\n");
  printf ("\n");
 
  /* exit the program */ 
  exit(1);
}



/* getArguments(argc, **argv, Arguments)
**
** This routine is sent the argument array from the command line and
** a Arguments struct used to store info from the command line.  It
** parses the arguments and stores the relevant information in the struct.
*/
void getArguments(int argc, char **argv, Arguments_t *Arguments) {
  int i = 0;      /* loop counter */
  int  err = 0;   /* flag for when an error occurs */

  /* check to see if we need to display the usage screen */
  if(argc == 1 || 
    (strcmp(argv[1], "-help") == 0)) { /* no command line args -> interact */
    usage();

  /* parse the arguments */
  } else {    
    /* initialize arguments */
    Arguments->origMP3FileName[0] = '\0';
    Arguments->newMP3FileName[0] = '\0';
    Arguments->hiddenFileName[0] = '\0';
    Arguments->hiding = 2;
    Arguments->bitSpacing = DEFAULT_SPACING;

    /* loop thorugh the arguments parsing each one one at a time */
    while(++i<argc && err == 0) {
      /* variabls for keeping track of arguments */
      char c, *token, *arg, *nextArg;
      int argUsed;
      token = argv[i];

      /* check the specified command flags */
      if(*token++ == '-') {
        if(i+1 < argc) {
          nextArg = argv[i+1];
        } else {
          nextArg = "";
        }
        argUsed = 0;
        /* loop through the flags */
        while(c = *token++) {
          if(*token) {
            arg = token;
          } else {
            arg = nextArg;
          }

          switch(c) {
            /* bit spacing was specified */
            case 'c': Arguments->bitSpacing = atoi(arg);
                      argUsed = 1;
                      break;

            /* hiding mode */    
            case 'r': Arguments->hiding = 0;
                      break;

            /* reading mode */
            case 'h': Arguments->hiding = 1;
                      break;

            /* verbose mode */
            case 'v': VERBOSE_MODE = 1;
                      break;
            
            /* error */
            default: fprintf(stderr,"ump3c: unrecognized option %c\n", c);
                     err = 1; 
                     break;
          }

          if(argUsed) {
            if(arg == token) {
              /* no more from token */
              token = ""; 
            } else {
              /* skip arg we used */
              ++i; 
            }
            /* reset variables */
            arg = ""; 
            argUsed = 0;
          }
        }

      /* read in the specified file names */
      } else {
        /* read in the first mp3 file name */
        /* if we are hiding: this is the mp3 that we create a new */
        /* mp3 from with the hidden message */
        /* if we are reading: this is the mp3 file that has the */
        /* hidden message in */
        if(Arguments->origMP3FileName[0] == '\0') {
          strcpy(Arguments->origMP3FileName, argv[i]);

        /* read in the new mp3 file name */
	/* this is the name of the newly created mp3 file with */
        /* the hidden message in (only used if we are hiding */
        } else if(Arguments->newMP3FileName[0] == '\0' &&
                  Arguments->hiding == 1) {
          strcpy(Arguments->newMP3FileName, argv[i]);

        /* read in the name of the hidden file that */
        /* (only used if hiding)*/
        } else if(Arguments->hiddenFileName[0] == '\0' &&
	          Arguments->hiding == 1) {
          strcpy(Arguments->hiddenFileName, argv[i]); 

        /* read in the name of the file to write the hidden */
        /* file to */
        } else if(Arguments->hiddenFileName[0] == '\0' &&
                  Arguments->hiding == 0) {
          strcpy(Arguments->hiddenFileName, argv[i]);
           
        /* too many args specified */
        } else {
          fprintf(stderr,"ump3c: excess arg %s\n", argv[i]);
          err = 1;
        }
      }
    }

    /* error check the command line results */
    if (err)  {
      printf ("Type \"ump3c -help\" for usage.\n");
      exit(1);
    }

    /* make sure an original mp3 file name was given */
    if (Arguments->origMP3FileName[0] == '\0')  {
      if (Arguments->hiding == 1) {
        printf("Need to specify name of original mp3 to hide file in.\n");
      } else {
        printf("Need to specify name of mp3 that contains the hidden file.\n");
      }
      printf("Type \"ump3c -help\" for usage.\n");
      exit(1);
    }

    /* make sure a new mp3 file name was given */
    if (Arguments->hiding == 1) {
      if (Arguments->newMP3FileName[0] == '\0') {
        printf("Need to specify new mp3 file name.\n");
        printf("Type \"ump3c -help\" for usage.\n");
        exit(1);
      }
      if (Arguments->hiddenFileName[0] == '\0' && 
          Arguments->hiding == 1) {
        printf("Need to specify the name of the file to be hidden.\n");
        printf("Type \"ump3c -help\" for usage.\n");
        exit(1);
      }
      if (Arguments->hiddenFileName[0] == '\0' && 
          Arguments->hiding == 0) {
        printf("Need to specify the name of the extracted hidden file.\n");
        printf("Type \"ump3c -help\" for usage.\n");
        exit(1);
      }
    }

    /* make sure the mode flag was set (hiding or reading */
    if (Arguments->hiding == 2) {
      printf ("Need to specify hiding(-h) or reading(-r).\n");
      printf("Type \"ump3c -help\" for usage.\n");
      exit(1); 
    }

    /* make sure if the spacing was specified that it is a positive value */
    if (Arguments->bitSpacing < 1) {
      printf("Spacing number(-c) needs to be larger than 0.\n");
      printf("Type \"ump3c -help\" for usage.\n");
      exit(1);
    }
  }
}



/* hideFileSize(*info)
**
** This routine is called from the III_put_side_info() routine in
** decode.c.  It is sent a ptr to a unsigned value that will have its
** least significant bit modified.  This routine is used to write the
** size of the file that is about to be hidden to the mp3.  It first
** writes a 5 bit header that will represent the number of bits used
** to represent the size. Followed by those bits to represent the size.
** The most significant bit will not be written since it is always 1.
** Since this routine is repeatedly called while iterating through the
** mp3, static variables are used to keep track of data from one
** call to the next and to mark which stage of writing the file size
** we are currently in. This routine returns 0 implying to the calling
** routine that this routine has more work to do otherwise it returns
** 1 implying that the hidden file size has been written to the mp3.
*/
int hideFileSize(unsigned * info) {
  int msb = 0;            /* most significant bit */
  static int nBits = 31;  /* num of bits needed to represent the file size */
                          /* starts at the 31st msb(i.e. 4 bytes for an int)*/
  static int nBitsAdjust = 0; /* adjusted copy of nBits */
  static int hideSizeStage = 1; /* flags which stage we are in */
  static int headerSize = 4;/* represents the 5 bit file size header */

  /* stage 1: figure out how many bits are needed to represent */
  /* the file szie. This is completely done this first time through */
  /* this routine */
  if (hideSizeStage == 1) {
    /* figure out where the msb is in the hidden file's size */
    while (msb != 1 && nBits >= 0){
      msb = (inFileSize & (1 << nBits)) ? 1 : 0;
      --nBits;
    }

    /* need to compensate for the last subtraction and 0 being the first */
    /* index */
    nBitsAdjust = nBits + 2;

    /* increment to the next stage and return 0 (not finished */
    ++hideSizeStage;
    return 0;

  /* stage 2: write the number of bits needed for file size into the */
  /* first 5 hidden bits */
  } else if (hideSizeStage == 2) {
    msb = (nBitsAdjust & (1 << headerSize)) ? 1 : 0;
    if(msb == 1) {
      *info = *info | 0x0001;
    } else {
      *info = *info & 0xfffe;
    }
    /* move to the next bit */
    --headerSize;

    /* check to see if we have written the entire 5 bit header */
    /* if so increment to the next stage and return not finished */
    if (headerSize < 0) {
      ++hideSizeStage;
    }
    return 0;

  /* stage 3: write the size to the next bits using as few bits as */
  /* necessary.  remeber we don't need to write the most sig bit since */
  /* it is always 1 */
  } else if (hideSizeStage == 3) {
    /* see what the current msb is */
    msb = (inFileSize & (1 << nBits)) ? 1 : 0;

    /* now make the lsb of the sent value the same as msb */
    if(msb == 1) {
      *info = *info | 0x0001;
    } else {
      *info = *info & 0xfffe;
    }
    /* move to the next bit */
    --nBits;

    /* check to see if we have hidden each bit of the header */
    /* if so increment to the next stage and return not finished */
    if (nBits < 0) {
      ++hideSizeStage;
    }
    return 0;

  /* we have completed all the stages so just return 1 to signal that */
  /* the hidden file size has already been hidden in the mp3 (finished) */
  } else {
    return 1;
  }
}



/* hideByte(*info)
**
** This routine is sent a ptr to a value and modifies
** it's least significant bit to the current bit in the file to
** be hidden. It reads in one byte at a time from the file to be hidden
** and writes one bit per call.
*/
int hideByte(unsigned * info) {
  char byte;    /* used to read in a byte of data from the file to be hidden */
  int msb = 0;  /* most significant bit */
  static int needNewByte = 1;   /* flags when we need to read in a new byte */
  static unsigned char currByte = 0; /* current byte that is being hidden */
  static int bitNum = 0;  /* marks which bit of the byte we are on */

  /* see if we need to read in a new byte */
  if (needNewByte) {
    /* read in the next byte of the file to be hidden */
    if (!(fread(&byte, sizeof(char),1, ptrHiddenFile))) {
      /* we reached the end of the file so return */
      return 1;
    }
    currByte = byte;

    /* changed to flag to mark that we don't need a new byte right now */
    needNewByte = 0;
  }

  /* now change the lsb of the sent value to match the currByte's bit */
  msb = (currByte & (1 << 7)) ? 1 : 0;
  if(msb == 1) {
    *info = *info | 0x0001;
  } else {
    *info = *info & 0xfffe;
  }

  /* advance to the next bit in the current byte */
  currByte = currByte << 1;
  ++bitNum;

  /* check to see if we need a new byte from the input file */
  if (bitNum == 8) {
    /* reset necessary variables */
    needNewByte = 1;
    bitNum = 0;
  
    /* increment counter to keep track of how many bytes have been hidden */
    ++nBytesHidden;
  }

  return 0;
}



/* readFileSize(info)
**
** This routine is sent a value and will retrieve its least significant
** bit to build the file size header and actual file size.  The header
** is a 5 bit number that represents the number of following bits that make
** up the hidden file size (in bytes). Since this routine is called 
** repeatedly while incrementing through the mp3 static variables were
** used to keep track of data and mark which stage of extracting the
** hidden file size we are currently in.  This routine returns 0 to signal
** the calling routine that it has more work to do, otherwise it returns
** 1 to signal that it is done.
*/
int readFileSize(unsigned info) {
  int lsb;                    /* least significant bit */
  static int headerSize = 5;  /* size in bits of file size header */
  static int nBits = 0;       /* number of bits that represent the hidden */
                              /* file size. This is the number represented */
                              /* in the 5 bit header */
  static int readSizeStage = 1;  /* flags which stage we are in */

  /* stage 1:we need to first read the file size header to see how many bytes */
  /* are used to represent the actual hidden file size */
  if (readSizeStage == 1) {
    /* get the least significant bit of the sent value */
    lsb = (info & (1)) ? 1 : 0;

    /* add it to the number of bits that represent the file size */
    nBits = nBits + lsb;

    /* move to the next bit in the header */
    --headerSize;

    /* check to see if we have read all of the header's bits */
    if (headerSize < 0) {
      /* zero out the most sig bit */
      nBits = nBits & 0xffdf;
      /* increment to the next stage */
      ++readSizeStage;
      return 0;
    } else {
      /* not done so shift over a bit to make room for the next */
      nBits = nBits << 1;
      return 0;
    }

  /* stage 2: now read in the actual size of the hidden file */
  /* nBits is the amount of bits we need to read in */
  } else if (readSizeStage == 2) {
    /* get the least significant bit of the sent value */
    lsb = (info & (1)) ? 1 : 0;

    /* add it to the number of bits that represent the file size */
    hiddenFileSize = hiddenFileSize + lsb;

    /* get to the next bit */
    --nBits;

    /* see if we have read in all the bits */
    /* need to adjust by 2 since we don't care about the msb and 0 is */
    /* our first bit */
    if ((nBits-2) < 0) {
      ++readSizeStage;
      return 0;
    } else {
      /* shift a bit to make room for the next */
      hiddenFileSize = hiddenFileSize << 1;
      return 0;
    }

  /* we have read in the hidden file size so return 1 to signal */
  /* that we are done */
  } else {
    return 1;
  }
}



/* readHiddenByte(info)
**
** This routine is sent a value that should contain one
** bit of the hidden message stored in it's least significant bit.
** It builds up a byte by one bit per call.  After it build an
** entire byte, the byte is written to the output file that 
** becomes the hidden file.
*/
int readHiddenByte(unsigned info) {
  int lsb;   /* least significant bit */
  static int needNewByte = 1;  /* flag that signals when we need a new byte */
  static unsigned char currByte = 0;  /* current byte */

  /* get the least significant bit */
  lsb = (info & (1)) ? 1 : 0;
  /* add it to the current byte that we are building */
  currByte = currByte + lsb;
  ++needNewByte;

  /* check to see if we have built a complete byte yet */
  if (needNewByte > 8) {
    /* see if we have read in all the bytes */
    if (nBytesRead == hiddenFileSize) {
      return 1;
    }

    /* write the current byte to the output file */
    fputc(currByte, ptrHiddenFile);
    /* increment counter that tells how many bytes have been read so far */
    ++nBytesRead;

    /* reset the variables for the next byte */
    needNewByte = 1;
    currByte = 0;

  /* not done building the current byte yet so shift over a bit */
  } else {
    currByte = currByte << 1;
  }

  return 0;
}




